@@ -220,7 +220,11 @@ def parse_dependencies(source_filename):
     # Actual parsing is way to slow, so we use regular expressions.
     # The only catch is that we must strip comments and string
     # literals ahead of time.
-    source = Utils.open_source_file(source_filename, "rU").read()
+    fh = Utils.open_source_file(source_filename, "rU")
+    try:
+        source = fh.read()
+    finally:
+        fh.close()
     distutils_info = DistutilsInfo(source)
     source, literals = strip_string_literals(source)
     source = source.replace('\\\n', ' ')
@@ -430,9 +434,12 @@ def create_extension_list(patterns, exclude=[], ctx=None, aliases=None):
                     for key, value in base.values.items():
                         if key not in kwds:
                             kwds[key] = value
+                sources = [file]
+                if template is not None:
+                    sources += template.sources[1:]
                 module_list.append(exn_type(
                         name=module_name,
-                        sources=[file],
+                        sources=sources,
                         **kwds))
                 m = module_list[-1]
                 seen.add(name)
